from collections import Counter
from flask import session
import cqi
import math
import random
from app import db, socketio
from app.decorators import socketio_login_required
from app.models import Corpus
from . import NAMESPACE as ns
from .utils import cqi_over_socketio, lookups_by_cpos


@socketio.on('cqi.corpora.corpus.drop', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_drop(cqi_client: cqi.CQiClient, corpus_name: str):
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.drop()
    payload = {'code': cqi_status,
               'msg': cqi.api.specification.lookup[cqi_status]}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.query', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_query(cqi_client: cqi.CQiClient, corpus_name: str, subcorpus_name: str, query: str):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.query(subcorpus_name, query)
    payload = {'code': cqi_status,
               'msg': cqi.api.specification.lookup[cqi_status]}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


###############################################################################
# nopaque specific CQi extensions                                             #
###############################################################################
@socketio.on('cqi.corpora.corpus.update_db', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_update_db(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = Corpus.query.get(session['d']['corpus_id'])
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    corpus.num_tokens = cqi_corpus.size
    db.session.commit()


@socketio.on('cqi.corpora.corpus.get_visualization_data', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_get_visualization_data(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = cqi_client.corpora.get(corpus_name)
    text = corpus.structural_attributes.get('text')
    s = corpus.structural_attributes.get('s')
    ent = corpus.structural_attributes.get('ent')
    word = corpus.positional_attributes.get('word')
    lemma = corpus.positional_attributes.get('lemma')
    pos = corpus.positional_attributes.get('pos')
    simple_pos = corpus.positional_attributes.get('simple_pos')
    payload = {}
    payload['corpus'] = {'lexicon': {}, 'values': []}
    payload['corpus']['lexicon'][0] = {
        'bounds': [0, corpus.size - 1],
        'counts': {
            'text': text.size,
            's': s.size,
            'ent': ent.size,
            'token': corpus.size
        },
        'freqs': {
            'word': dict(
                zip(
                    range(0, word.lexicon_size),
                    word.freqs_by_ids(list(range(0, word.lexicon_size)))
                )
            ),
            'lemma': dict(
                zip(
                    range(0, lemma.lexicon_size),
                    lemma.freqs_by_ids(list(range(0, lemma.lexicon_size)))
                )
            ),
            'pos': dict(
                zip(
                    range(0, pos.lexicon_size),
                    pos.freqs_by_ids(list(range(0, pos.lexicon_size)))
                )
            ),
            'simple_pos': dict(
                zip(
                    range(0, simple_pos.lexicon_size),
                    simple_pos.freqs_by_ids(list(range(0, simple_pos.lexicon_size)))
                )
            )
        }
    }
    payload['text'] = {'lexicon': {}, 'values': None}
    for text_id in range(0, text.size):
        text_lbound, text_rbound = text.cpos_by_id(text_id)
        text_cpos_range = range(text_lbound, text_rbound + 1)
        text_s_ids = s.ids_by_cpos(list(text_cpos_range))
        text_ent_ids = ent.ids_by_cpos(list(text_cpos_range))
        payload['text']['lexicon'][text_id] = {
            'bounds': [text_lbound, text_rbound],
            'counts': {
                's': len([x for x in text_s_ids if x != -1]),
                'ent': len([x for x in text_ent_ids if x != -1]),
                'token': text_rbound - text_lbound + 1
            },
            'freqs': {
                'word': dict(
                    Counter(word.ids_by_cpos(list(text_cpos_range)))
                ),
                'lemma': dict(
                    Counter(lemma.ids_by_cpos(list(text_cpos_range)))
                ),
                'pos': dict(
                    Counter(pos.ids_by_cpos(list(text_cpos_range)))
                ),
                'simple_pos': dict(
                    Counter(simple_pos.ids_by_cpos(list(text_cpos_range)))
                )
            }
        }
    payload['text']['values'] = [
        sub_attr.name[(len(text.name) + 1):]
        for sub_attr in corpus.structural_attributes.list(filters={'part_of': text})
    ]
    payload['s'] = {'lexicon': {}, 'values': None}
    for s_id in range(0, s.size):
        payload['s']['lexicon'][s_id] = {
            # 'bounds': s.cpos_by_id(s_id)
        }
    payload['s']['values'] = [
        sub_attr.name[(len(s.name) + 1):]
        for sub_attr in corpus.structural_attributes.list(filters={'part_of': s})
    ]
    payload['ent'] = {'lexicon': {}, 'values': None}
    for ent_id in range(0, ent.size):
        payload['ent']['lexicon'][ent_id] = {
            # 'bounds': ent.cpos_by_id(ent_id)
        }
    payload['ent']['values'] = [
        sub_attr.name[(len(ent.name) + 1):]
        for sub_attr in corpus.structural_attributes.list(filters={'part_of': ent})
    ]
    payload['lookups'] = {
        'corpus': {},
        'text': {},
        's': {},
        'ent': {},
        'word': dict(
            zip(
                range(0, word.lexicon_size),
                word.values_by_ids(list(range(0, word.lexicon_size)))
            )
        ),
        'lemma': dict(
            zip(
                range(0, lemma.lexicon_size),
                lemma.values_by_ids(list(range(0, lemma.lexicon_size)))
            )
        ),
        'pos': dict(
            zip(
                range(0, pos.lexicon_size),
                pos.values_by_ids(list(range(0, pos.lexicon_size)))
            )
        ),
        'simple_pos': dict(
            zip(
                range(0, simple_pos.lexicon_size),
                simple_pos.values_by_ids(list(range(0, simple_pos.lexicon_size)))
            )
        )
    }
    # print(payload)
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.paginate', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_paginate(cqi_client: cqi.CQiClient, corpus_name: str, page: int = 1, per_page: int = 20):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_corpus.size > 0
            and page > math.ceil(cqi_corpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    first_cpos = (page - 1) * per_page
    last_cpos = min(cqi_corpus.size, first_cpos + per_page)
    cpos_list = [*range(first_cpos, last_cpos)]
    lookups = lookups_by_cpos(cqi_corpus, cpos_list)
    payload = {}
    # the items for the current page
    payload['items'] = [cpos_list]
    # the lookups for the items
    payload['lookups'] = lookups
    # the total number of items matching the query
    payload['total'] = cqi_corpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return {'code': 200, 'msg': 'OK', 'payload': payload}
